import asyncio
import configparser
import logging
import argparse
import os.path

from smart_tv_telegram import *


def open_config(parser: argparse.ArgumentParser, arg: str) -> Config:
    if not os.path.exists(arg):
        parser.error(f"The file `{arg}` does not exist")

    elif not os.path.isfile(arg):
        parser.error(f"`{arg}` is not a file")

    try:
        return Config(arg)
    except ValueError as err:
        parser.error(str(err))
    except KeyError as err:
        parser.error(f"config key {str(err)} does not exists")
    except configparser.Error as err:
        parser.error(f"generic configparser error:\n{str(err)}")


async def async_main(config: Config):
    mtproto = Mtproto(config)
    http = Http(mtproto, config)

    bot = Bot(mtproto, config)
    bot.prepare()

    await mtproto.start()
    await http.start()


def main(config: Config):
    loop = asyncio.get_event_loop()
    loop.run_until_complete(async_main(config))


def arg_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config", type=lambda x: open_config(parser, x), default="config.ini")
    parser.add_argument("-v", "--verbosity", type=int, choices=[0, 1, 2], default=0)

    args = parser.parse_args()

    if args.verbosity == 0:
        logging.basicConfig(level=logging.ERROR)

    elif args.verbosity == 1:
        logging.basicConfig(level=logging.INFO)

    elif args.verbosity == 2:
        logging.basicConfig(level=logging.DEBUG)

    main(args.config)


if __name__ == "__main__":
    arg_parser()



import typing
from urllib.parse import quote

import aiohttp.web
from aiohttp import web
from aiohttp.web_request import Request
from aiohttp.web_response import Response, StreamResponse
from pyrogram.raw.types import MessageMediaDocument, Document

from . import Config, Mtproto
from .tools import parse_http_range, mtproto_filename


__all__ = [
    "Http"
]


class Http:
    _mtproto: Mtproto
    _config: Config

    def __init__(self, mtproto: Mtproto, config: Config):
        self._mtproto = mtproto
        self._config = config

    async def start(self):
        app = web.Application()
        app.add_routes([web.get("/stream/{message_id}", self._stream_handler)])
        app.add_routes([web.options("/stream/{message_id}", self._upnp_discovery_handler)])
        app.add_routes([web.put("/stream/{message_id}", self._upnp_discovery_handler)])
        app.add_routes([web.get("/healthcheck", self._health_check_handler)])

        # noinspection PyProtectedMember
        await aiohttp.web._run_app(app, host=self._config.listen_host, port=self._config.listen_port)

    def _write_upnp_headers(self, result: typing.Union[Response, StreamResponse]):
        result.headers.setdefault("Content-Type", "video/mp4")
        result.headers.setdefault("Access-Control-Allow-Origin", "*")
        result.headers.setdefault("Access-Control-Allow-Methods", "GET, OPTIONS")
        result.headers.setdefault("Access-Control-Allow-Headers", "Content-Type")
        result.headers.setdefault("transferMode.dlna.org", "Streaming")
        result.headers.setdefault("TimeSeekRange.dlna.org", "npt=0.00-")
        result.headers.setdefault("contentFeatures.dlna.org", "DLNA.ORG_OP=01;DLNA.ORG_CI=0;")

    def _write_filename_header(self, result: typing.Union[Response, StreamResponse], filename: str):
        result.headers.setdefault("Content-Disposition", f'inline; filename="{quote(filename)}"')

    async def _health_check_handler(self, _: Request) -> typing.Optional[Response]:
        try:
            await self._mtproto.health_check()
            return Response(status=200, text="ok")
        except ConnectionError:
            return Response(status=500, text="gone")

    async def _upnp_discovery_handler(self, _: Request) -> typing.Optional[Response]:
        result = Response(status=200)
        self._write_upnp_headers(result)
        return result

    async def _stream_handler(self, request: Request) -> typing.Optional[Response]:
        message_id: str = request.match_info["message_id"]

        if not message_id.isdigit():
            return Response(status=401)

        range_header = request.headers.get("Range")

        if range_header is None:
            offset = 0
            data_to_skip = False
            max_size = None

        else:
            try:
                offset, data_to_skip, max_size = parse_http_range(
                    range_header, self._config.block_size)
            except ValueError:
                return Response(status=400)

        if data_to_skip > self._config.block_size:
            return Response(status=500)

        try:
            message = await self._mtproto.get_message(int(message_id))
        except ValueError:
            return Response(status=404)

        if not isinstance(message.media, MessageMediaDocument):
            return Response(status=404)

        if not isinstance(message.media.document, Document):
            return Response(status=404)

        size = message.media.document.size
        read_after = offset + data_to_skip

        if read_after > size:
            return Response(status=400)

        if (max_size is not None) and (size < max_size):
            return Response(status=400)

        if max_size is None:
            max_size = size

        stream = StreamResponse(status=206 if (read_after or (max_size != size)) else 200)
        self._write_upnp_headers(stream)

        stream.headers.setdefault("Content-Range", f"bytes {read_after}-{max_size}/{size}")
        stream.headers.setdefault("Accept-Ranges", "bytes")
        stream.headers.setdefault("Content-Length", str(size))

        try:
            filename = mtproto_filename(message)
        except TypeError:
            filename = f"file_{message.media.document.id}"

        self._write_filename_header(stream, filename)

        await stream.prepare(request)

        while offset < max_size:
            block = await self._mtproto.get_block(message, offset, self._config.block_size)
            new_offset = offset + len(block)

            if data_to_skip:
                block = block[data_to_skip:]
                data_to_skip = False

            if new_offset > max_size:
                block = block[:-(new_offset - max_size)]

            offset = new_offset

            if request.transport.is_closing():
                break

            await stream.write(block)

        await stream.write_eof()



import asyncio
import concurrent.futures
import re
import typing

from pyrogram.raw.types import MessageMediaDocument, Document, DocumentAttributeFilename
from pyrogram.raw.types import Message as TlMessage
from pyrogram.types import Message as BoxedMessage

from . import Config


__all__ = [
    "mtproto_filename",
    "build_uri",
    "ascii_only",
    "run_method_in_executor",
    "parse_http_range",
    "pyrogram_filename"
]


_NAMED_MEDIA_TYPES = ("document", "video", "audio", "video_note", "animation")
_RANGE_REGEX = re.compile(r"bytes=([0-9]+)-([0-9]+)?")
_EXECUTOR = concurrent.futures.ThreadPoolExecutor()
_LOOP = asyncio.get_event_loop()


def pyrogram_filename(message: BoxedMessage) -> str:
    try:
        return next(
            getattr(message, t).file_name
            for t in _NAMED_MEDIA_TYPES
            if getattr(message, t) is not None
        )
    except StopIteration:
        raise TypeError()


def mtproto_filename(message: TlMessage) -> str:
    if not (
        isinstance(message.media, MessageMediaDocument) and
        isinstance(message.media.document, Document)
    ):
        raise TypeError()

    try:
        return next(
            attr.file_name
            for attr in message.media.document.attributes
            if isinstance(attr, DocumentAttributeFilename)
        )
    except StopIteration:
        raise TypeError()


def build_uri(config: Config, msg_id: int) -> str:
    return f"http://{config.listen_host}:{config.listen_port}/stream/{msg_id}"


def ascii_only(haystack: str) -> str:
    return "".join(c for c in haystack if ord(c) < 128)


def run_method_in_executor(func):
    async def wraps(*args):
        return await _LOOP.run_in_executor(_EXECUTOR, func, *args)
    return wraps


def parse_http_range(http_range: str, block_size: int) -> typing.Tuple[int, int, typing.Optional[int]]:
    matches = _RANGE_REGEX.search(http_range)

    if matches is None:
        raise ValueError()

    offset = matches.group(1)

    if not offset.isdigit():
        raise ValueError()

    max_size = matches.group(2)

    if max_size and max_size.isdigit():
        max_size = int(max_size)
    else:
        max_size = None

    offset = int(offset)
    safe_offset = (offset // block_size) * block_size
    data_to_skip = offset - safe_offset

    return safe_offset, data_to_skip, max_size



import asyncio
import io
import logging
import typing

from . import DeviceFinder, Device
from .. import Config


__all__ = [
    "VlcDevice",
    "VlcDeviceFinder"
]


_LOGGER = logging.getLogger(__name__)
_ENCODING = "utf8"
_EOF = b"\n\r"
_AUTH_MAGIC = b"\xff\xfb\x01"
_AUTH_OK = b"\xff\xfc\x01\r\nWelcome"


class VlcDeviceParams:
    _host: str
    _port: int
    _password: typing.Optional[str] = None

    def __init__(self, params: typing.Dict[str, typing.Any]):
        self._host = params["host"]
        self._port = params["port"]

        if "password" in params:
            self._password = params["password"]

    @property
    def host(self) -> str:
        return self._host

    @property
    def port(self) -> int:
        return self._port

    @property
    def password(self) -> typing.Optional[str]:
        return self._password


class VlcDevice(Device):
    params: VlcDeviceParams

    # noinspection PyMissingConstructor
    def __init__(self, device: VlcDeviceParams):
        self._params = device

    def get_device_name(self) -> str:
        return f"vlc @{self._params.host}"

    async def _call(self, method: str, *args: typing.AnyStr):
        reader, writer = await asyncio.open_connection(self._params.host, self._params.port)
        headers = await reader.read(io.DEFAULT_BUFFER_SIZE)

        if headers.endswith(_AUTH_MAGIC):
            if self._params.password:
                writer.write(bytes(self._params.password, _ENCODING) + _EOF)
                await writer.drain()

                auth_result = await reader.read(io.DEFAULT_BUFFER_SIZE)

                if not auth_result.startswith(_AUTH_OK):
                    _LOGGER.error("receive: %s", auth_result.decode(_ENCODING, "ignore"))
                    return writer.close()

            else:
                _LOGGER.error("vlc %s: need password", self._params.host)
                return writer.close()

        writer.write(
            method.encode(_ENCODING) + b" " +
            b" ".join(a.encode(_ENCODING) for a in args) + _EOF
        )

        await writer.drain()
        return writer.close()

    async def stop(self):
        await self._call("stop")

    async def play(self, url: str, title: str):
        await self._call("add", url)
        await self._call("play")


class VlcDeviceFinder(DeviceFinder):
    @staticmethod
    async def find(config: Config) -> typing.List[Device]:
        return [
            VlcDevice(VlcDeviceParams(params))
            for params in config.vlc_devices
        ]



import abc
import typing

from .. import Config


__all__ = [
    "Device",
    "DeviceFinder"
]


class Device(abc.ABC):
    # noinspection PyUnusedLocal
    @abc.abstractmethod
    def __init__(self, device: typing.Any):
        raise NotImplementedError

    @abc.abstractmethod
    async def stop(self):
        raise NotImplementedError

    @abc.abstractmethod
    async def play(self, url: str, title: str):
        raise NotImplementedError

    @abc.abstractmethod
    def get_device_name(self) -> str:
        raise NotImplementedError

    def __repr__(self):
        return self.get_device_name()


class DeviceFinder(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    async def find(config: Config) -> typing.List[Device]:
        raise NotImplementedError



from .config import Config
from .mtproto import Mtproto
from .http import Http
from .bot import Bot


__version__ = "1.0.1"
__version_info__ = ("1", "0", "1")
__author__ = "https://github.com/andrew-ld"


__all__ = [
    "Config",
    "Mtproto",
    "Http",
    "Bot",
    "__version__",
    "__version_info__",
    "__author__",
]



import abc
import functools
import typing
import enum
import html

import async_timeout
from pyrogram import Client, filters
from pyrogram.filters import create
from pyrogram.handlers import MessageHandler
from pyrogram.types import ReplyKeyboardRemove, Message, KeyboardButton, ReplyKeyboardMarkup

from . import Config, Mtproto
from .devices import UpnpDeviceFinder, ChromecastDeviceFinder, VlcDeviceFinder, XbmcDeviceFinder, Device
from .tools import build_uri, pyrogram_filename


__all__ = [
    "Bot"
]


_REMOVE_KEYBOARD = ReplyKeyboardRemove()
_CANCEL_BUTTON = "^Cancel"


class States(enum.Enum):
    NOTHING = enum.auto()
    SELECT = enum.auto()


class StateData(abc.ABC):
    @abc.abstractmethod
    def get_associated_state(self) -> States:
        raise NotImplementedError


class SelectStateData(StateData):
    msg_id: int
    filename: str
    devices: typing.List[Device]

    def get_associated_state(self) -> States:
        return States.SELECT

    def __init__(self, msg_id: int, filename: str, devices: typing.List[Device]):
        self.msg_id = msg_id
        self.filename = filename
        self.devices = devices


class TelegramStateMachine:
    _states: typing.Dict[int, typing.Tuple[States, typing.Union[bool, StateData]]]

    def __init__(self):
        self._states = {}

    def get_state(self, message: Message) -> typing.Tuple[States, typing.Union[bool, StateData]]:
        user_id = message.from_user.id

        if user_id in self._states:
            return self._states[user_id]

        return States.NOTHING, False

    def set_state(self, message: Message, state: States, data: typing.Union[bool, StateData]) -> bool:
        if isinstance(data, bool) or data.get_associated_state() == state:
            self._states[message.from_user.id] = (state, data)
            return True

        raise TypeError()


class Bot:
    _config: Config
    _state_machine: TelegramStateMachine
    _mtproto: Mtproto

    def __init__(self, mtproto: Mtproto, config: Config):
        self._config = config
        self._mtproto = mtproto
        self._state_machine = TelegramStateMachine()

    def prepare(self):
        admin_filter = filters.chat(self._config.admins) & filters.private
        self._mtproto.register(MessageHandler(self._new_document, filters.document & admin_filter))
        self._mtproto.register(MessageHandler(self._new_document, filters.video & admin_filter))
        self._mtproto.register(MessageHandler(self._new_document, filters.audio & admin_filter))
        self._mtproto.register(MessageHandler(self._new_document, filters.animation & admin_filter))
        self._mtproto.register(MessageHandler(self._new_document, filters.voice & admin_filter))
        self._mtproto.register(MessageHandler(self._new_document, filters.video_note & admin_filter))

        state_filter = create(lambda _, __, m: self._state_machine.get_state(m)[0] == States.SELECT)
        self._mtproto.register(MessageHandler(self._select_device, filters.text & admin_filter & state_filter))

    async def _select_device(self, _: Client, message: Message):
        data: SelectStateData
        _, data = self._state_machine.get_state(message)

        self._state_machine.set_state(message, States.NOTHING, False)
        reply = functools.partial(message.reply, reply_markup=_REMOVE_KEYBOARD)

        if message.text == _CANCEL_BUTTON:
            await reply("Cancelled")
            return

        try:
            device = next(
                device
                for device in data.devices
                if repr(device) == message.text
            )
        except StopIteration:
            await reply("Wrong device")
            return

        async with async_timeout.timeout(self._config.device_request_timeout) as cm:
            uri = build_uri(self._config, data.msg_id)

            # noinspection PyBroadException
            try:
                await device.stop()
                await device.play(uri, data.filename)

            except Exception as ex:
                await reply(
                    "Error while communicate with the device:\n\n"
                    f"<code>{html.escape(str(ex))}</code>"
                )

            else:
                await reply(f"Playing ID: {data.msg_id}")

        if cm.expired:
            await reply(f"Timeout while communicate with the device")

    async def _new_document(self, _: Client, message: Message):
        devices = []

        if self._config.upnp_enabled:
            devices.extend(await UpnpDeviceFinder.find(self._config))

        if self._config.chromecast_enabled:
            # noinspection PyUnresolvedReferences
            devices.extend(await ChromecastDeviceFinder.find(self._config))

        if self._config.xbmc_enabled:
            devices.extend(await XbmcDeviceFinder.find(self._config))

        if self._config.vlc_enabled:
            devices.extend(await VlcDeviceFinder.find(self._config))

        if devices:
            try:
                filename = pyrogram_filename(message)
            except TypeError:
                filename = "NaN"

            self._state_machine.set_state(
                message,
                States.SELECT,
                SelectStateData(
                    message.message_id,
                    filename,
                    devices.copy()
                )
            )

            buttons = [[KeyboardButton(repr(device))] for device in devices]
            buttons.append([KeyboardButton(_CANCEL_BUTTON)])
            markup = ReplyKeyboardMarkup(buttons, one_time_keyboard=True)
            await message.reply("Select a device", reply_markup=markup)

        else:
            await message.reply("Supported devices not found in the network", reply_markup=_REMOVE_KEYBOARD)



import asyncio
import functools
import os
import pickle

import pyrogram

from async_lru import alru_cache
from pyrogram.handlers.handler import Handler
from pyrogram.raw.functions.auth import ExportAuthorization, ImportAuthorization
from pyrogram.raw.functions.help import GetConfig
from pyrogram.raw.functions.messages import GetMessages
from pyrogram.raw.functions.upload import GetFile
from pyrogram.raw.types import InputMessageID, Message, InputDocumentFileLocation
from pyrogram.errors import FloodWait
from pyrogram.session import Session

from . import Config


__all__ = [
    "Mtproto"
]


class Mtproto:
    _config: Config
    _client: pyrogram.Client

    def __init__(self, config: Config):
        self._config = config
        self._client = pyrogram.Client(config.session_name, config.api_id, config.api_hash,
                                       bot_token=config.token, sleep_threshold=0, workdir=os.getcwd())

    def register(self, handler: Handler):
        self._client.add_handler(handler)

    @alru_cache(cache_exceptions=False)
    async def get_message(self, message_id: int) -> Message:
        messages = await self._client.send(GetMessages(id=[InputMessageID(id=message_id)]))

        if not messages.messages:
            raise ValueError("wrong message_id")

        message = messages.messages[0]

        if not isinstance(message, Message):
            raise ValueError(f"expected `Message`, found: `{type(message).__name__}`")

        return message

    async def health_check(self):
        if not all(
            session.is_connected.is_set()
            for session in (
                *self._client.media_sessions.values(),
                self._client.session
            )
        ):
            raise ConnectionError()

    async def get_block(self, message: Message, offset: int, block_size: int) -> bytes:
        session = self._client.media_sessions.get(message.media.document.dc_id)

        r = GetFile(
            offset=offset,
            limit=block_size,
            location=InputDocumentFileLocation(
                id=message.media.document.id,
                access_hash=message.media.document.access_hash,
                file_reference=b"",
                thumb_size=""
            )
        )

        while True:
            try:
                result = await session.send(r, sleep_threshold=0)
            except FloodWait:  # file floodwait is fake
                await asyncio.sleep(self._config.file_fake_fw_wait)
            else:
                break

        return result.bytes

    async def start(self):
        await self._client.start()

        config = await self._client.send(GetConfig())
        dc_ids = [x.id for x in config.dc_options]
        keys_path = self._config.session_name + ".keys"

        if os.path.exists(keys_path):
            keys = pickle.load(open(keys_path, "rb"))
        else:
            keys = {}

        for dc_id in dc_ids:
            session = functools.partial(pyrogram.session.Session, self._client, dc_id, is_media=True, test_mode=False)

            if dc_id != await self._client.storage.dc_id():
                if dc_id not in keys:
                    exported_auth = await self._client.send(ExportAuthorization(dc_id=dc_id))

                    auth = pyrogram.session.Auth(self._client, dc_id, False)
                    auth_key = await auth.create()

                    session = session(auth_key)
                    await session.start()

                    await session.send(ImportAuthorization(id=exported_auth.id, bytes=exported_auth.bytes))
                    keys[dc_id] = session.auth_key

                else:
                    session = session(keys[dc_id])
                    await session.start()

            else:
                session = session(await self._client.storage.auth_key())
                await session.start()

            self._client.media_sessions[dc_id] = session

        pickle.dump(keys, open(keys_path, "wb"))



# from: https://github.com/home-assistant/core/blob/dev/homeassistant/components/kodi/media_player.py

import asyncio
import json
import logging
import uuid
import aiohttp
import typing

from .. import Config
from . import Device, DeviceFinder


__all__ = [
    "XbmcDevice",
    "XbmcDeviceFinder"
]


_LOGGER = logging.getLogger(__name__)
_ARG_TYPE = typing.Union[typing.AnyStr, int, bool]

_JSON_HEADERS = {"content-type": "application/json"}
_JSONRPC_VERSION = "2.0"

_ATTR_JSONRPC = "jsonrpc"
_ATTR_METHOD = "method"
_ATTR_PARAMS = "params"
_ATTR_ID = "id"


class XbmcDeviceParams:
    _host: str
    _port: int
    _username: typing.Optional[str] = None
    _password: typing.Optional[str] = None

    def __init__(self, params: typing.Dict[str, typing.AnyStr]):
        self._host = params["host"]
        self._port = params["port"]

        if "username" in params:
            self._username = params["username"]
            self._password = params["password"]

    @property
    def host(self) -> str:
        return self._host

    @property
    def port(self) -> int:
        return self._port

    @property
    def username(self) -> typing.Optional[str]:
        return self._username

    @property
    def password(self) -> typing.Optional[str]:
        return self._password


class XbmcDevice(Device):
    _auth: typing.Optional[aiohttp.BasicAuth]
    _http_url: str
    _host: str

    # noinspection PyMissingConstructor
    def __init__(self, device: XbmcDeviceParams):
        if device.username:
            self._auth = aiohttp.BasicAuth(device.username, device.password)
        else:
            self._auth = None

        self._http_url = f"http://{device.host}:{device.port}/jsonrpc"
        self._host = device.host

    def get_device_name(self) -> str:
        return f"xbmc @{self._host}"

    async def _call(self, method: str, **args: typing.Union[_ARG_TYPE, typing.Mapping[str, _ARG_TYPE]]):
        data = {
            _ATTR_JSONRPC: _JSONRPC_VERSION,
            _ATTR_METHOD: method,
            _ATTR_ID: str(uuid.uuid4()),
            _ATTR_PARAMS: args
        }

        response = None
        session = aiohttp.ClientSession(auth=self._auth, headers=_JSON_HEADERS)

        try:
            response = await session.post(self._http_url, data=json.dumps(data))

            if response.status == 401:
                _LOGGER.error(
                    "Error fetching Kodi data. HTTP %d Unauthorized. "
                    "Password is incorrect.", response.status)
                return None

            if response.status != 200:
                _LOGGER.error(
                    "Error fetching Kodi data. HTTP %d", response.status)
                return None

            response_json = await response.json()

            if "error" in response_json:
                _LOGGER.error(
                    "RPC Error Code %d: %s",
                    response_json["error"]["code"],
                    response_json["error"]["message"])
                return None

            return response_json["result"]

        except (aiohttp.ClientError,
                asyncio.TimeoutError,
                ConnectionRefusedError):
            return None

        finally:
            if response:
                response.close()

            await session.close()

    async def stop(self):
        players = await self._call("Player.GetActivePlayers")

        if players:
            await self._call("Player.Stop", playerid=players[0]["playerid"])

    async def play(self, url: str, title: str):
        await self._call("Playlist.Clear", playlistid=0)
        await self._call("Playlist.Add", playlistid=0, item={"file": url})
        await self._call("Player.Open", item={"playlistid": 0}, options={"repeat": "one"})


class XbmcDeviceFinder(DeviceFinder):
    @staticmethod
    async def find(config: Config) -> typing.List[Device]:
        return [
            XbmcDevice(XbmcDeviceParams(params))
            for params in config.xbmc_devices
        ]



import typing
from ipaddress import IPv4Address
from xml.sax.saxutils import escape

import async_upnp_client
from async_upnp_client import UpnpFactory, UpnpError
from async_upnp_client.aiohttp import AiohttpRequester
from async_upnp_client.search import async_search

from . import Device, DeviceFinder
from .. import Config
from ..tools import ascii_only


__all__ = [
    "UpnpDevice",
    "UpnpDeviceFinder"
]


_AVTRANSPORT_SCHEMA = "urn:schemas-upnp-org:service:AVTransport:1"

_DLL_METADATA = """
<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/"
    xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
    xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/"
    xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">
    <item id="R:0/0/0" parentID="R:0/0" restricted="true">
        <dc:title>{title}</dc:title>
        <upnp:class>object.item.videoItem.movie</upnp:class>
        <desc id="cdudn" nameSpace="urn:schemas-rinconnetworks-com:metadata-1-0/">
            SA_RINCON65031_
        </desc>
        <res protocolInfo="http-get:*:video/mp4:DLNA.ORG_OP=01;DLNA.ORG_CI=0;DLNA.ORG_FLAGS=01700000000000000000000000000000">{url}</res>
    </item>
</DIDL-Lite>
"""


class UpnpDevice(Device):
    _device: async_upnp_client.UpnpDevice
    _service: async_upnp_client.UpnpService

    # noinspection PyMissingConstructor
    def __init__(self, device: async_upnp_client.UpnpDevice):
        self._device = device
        self._service = self._device.service(_AVTRANSPORT_SCHEMA)

    def get_device_name(self) -> str:
        return self._device.friendly_name

    async def stop(self):
        stop = self._service.action("Stop")

        try:
            await stop.async_call(InstanceID=0)
        except UpnpError as e:
            if "Transition not available" not in str(e):
                raise e

    async def play(self, url: str, title: str):
        set_url = self._service.action("SetAVTransportURI")
        meta = _DLL_METADATA.format(title=escape(ascii_only(title)), url=escape(url))
        await set_url.async_call(InstanceID=0, CurrentURI=url, CurrentURIMetaData=meta)

        play = self._service.action("Play")
        await play.async_call(InstanceID=0, Speed="1")


class UpnpDeviceFinder(DeviceFinder):
    @staticmethod
    async def find(config: Config) -> typing.List[Device]:
        devices = []
        requester = AiohttpRequester()
        factory = UpnpFactory(requester)
        source_ip = IPv4Address("0.0.0.0")

        async def on_response(data: typing.Mapping[str, typing.Any]) -> None:
            devices.append(await factory.async_create_device(data.get("LOCATION")))

        await async_search(service_type=_AVTRANSPORT_SCHEMA,
                           source_ip=source_ip,
                           timeout=config.upnp_scan_timeout,
                           async_callback=on_response)

        return [UpnpDevice(device) for device in devices]



import ast
import configparser
import typing


__all__ = [
    "Config"
]


class Config:
    _api_id: int
    _api_hash: str
    _token: str
    _session_name: str
    _file_fake_fw_wait: float

    _device_request_timeout: int

    _listen_host: str
    _listen_port: int

    _upnp_enabled: bool
    _upnp_scan_timeout: int = 0

    _chromecast_enabled: bool
    _chromecast_scan_timeout: int = 0

    _xbmc_enabled: bool
    _xbmc_devices: typing.List[dict]

    _vlc_enabled: bool
    _vlc_devices: typing.List[dict]

    _admins: typing.List[int]
    _block_size: int

    def __init__(self, path: str):
        config = configparser.ConfigParser()
        config.read(path)

        self._api_id = int(config["mtproto"]["api_id"])
        self._api_hash = str(config["mtproto"]["api_hash"])
        self._token = str(config["mtproto"]["token"])
        self._session_name = str(config["mtproto"]["session_name"])
        self._file_fake_fw_wait = float(config["mtproto"]["file_fake_fw_wait"])

        self._listen_port = int(config["http"]["listen_port"])
        self._listen_host = str(config["http"]["listen_host"])

        self._upnp_enabled = bool(int(config["discovery"]["upnp_enabled"]))

        if self._upnp_enabled:
            self._upnp_scan_timeout = int(config["discovery"]["upnp_scan_timeout"])

        self._chromecast_enabled = bool(int(config["discovery"]["chromecast_enabled"]))

        self._device_request_timeout = int(config["discovery"]["device_request_timeout"])

        self._xbmc_enabled = bool(int(config["discovery"]["xbmc_enabled"]))

        if self._xbmc_enabled:
            self._xbmc_devices = ast.literal_eval(config["discovery"]["xbmc_devices"])

            if not isinstance(self._xbmc_devices, list):
                raise ValueError("xbmc_devices should be a list")

            if not all(isinstance(x, dict) for x in self._xbmc_devices):
                raise ValueError("xbmc_devices should contain only dict")

        else:
            self._xbmc_devices = []

        self._vlc_enabled = bool(int(config["discovery"]["vlc_enabled"]))

        if self._vlc_enabled:
            self._vlc_devices = ast.literal_eval(config["discovery"]["vlc_devices"])

            if not isinstance(self._xbmc_devices, list):
                raise ValueError("vlc_devices should be a list")

            if not all(isinstance(x, dict) for x in self._xbmc_devices):
                raise ValueError("vlc_devices should contain only dict")

        else:
            self._vlc_devices = []

        if self._chromecast_enabled:
            self._chromecast_scan_timeout = int(config["discovery"]["chromecast_scan_timeout"])

        self._admins = ast.literal_eval(config["bot"]["admins"])
        self._block_size = int(config["bot"]["block_size"])

        if not isinstance(self._admins, list):
            raise ValueError("admins should be a list")

        if not all(isinstance(x, int) for x in self._admins):
            raise ValueError("admins list should contain only integers")

    @property
    def file_fake_fw_wait(self) -> float:
        return self._file_fake_fw_wait

    @property
    def api_id(self) -> int:
        return self._api_id

    @property
    def api_hash(self) -> str:
        return self._api_hash

    @property
    def token(self) -> str:
        return self._token

    @property
    def session_name(self) -> str:
        return self._session_name

    @property
    def listen_host(self) -> str:
        return self._listen_host

    @property
    def listen_port(self) -> int:
        return self._listen_port

    @property
    def upnp_enabled(self) -> bool:
        return self._upnp_enabled

    @property
    def upnp_scan_timeout(self) -> int:
        return self._upnp_scan_timeout

    @property
    def chromecast_enabled(self) -> bool:
        return self._chromecast_enabled

    @property
    def chromecast_scan_timeout(self) -> int:
        return self._chromecast_scan_timeout

    @property
    def xbmc_enabled(self) -> bool:
        return self._xbmc_enabled

    @property
    def xbmc_devices(self) -> typing.List[dict]:
        return self._xbmc_devices

    @property
    def vlc_enabled(self) -> bool:
        return self._vlc_enabled

    @property
    def vlc_devices(self) -> typing.List[dict]:
        return self._vlc_devices

    @property
    def admins(self) -> typing.List[int]:
        return self._admins

    @property
    def block_size(self) -> int:
        return self._block_size

    @property
    def device_request_timeout(self) -> int:
        return self._device_request_timeout



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()


with open("requirements.txt") as fh:
    requirements = fh.read().splitlines()


setuptools.setup(
    name="smart_tv_telegram",
    version="1.0.1.dev0",
    author="andrew-ld",
    author_email="andrew-ld@protonmail.com",
    description="A Telegram Bot to stream content on your smart TV",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/andrew-ld/smart-tv-telegram",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    python_requires=">=3.6",
    entry_points={
        "console_scripts": [
            "smart_tv_telegram=smart_tv_telegram.__main__:arg_parser"
        ],
    }
)



from .device import Device, DeviceFinder
from .upnp_device import UpnpDevice, UpnpDeviceFinder
from .chromecast_device import ChromecastDevice, ChromecastDeviceFinder
from .vlc_device import VlcDeviceFinder, VlcDevice
from .xbmc_device import XbmcDevice, XbmcDeviceFinder


__all__ = [
    "Device",
    "DeviceFinder",
    "UpnpDevice",
    "UpnpDeviceFinder",
    "ChromecastDevice",
    "ChromecastDeviceFinder",
    "XbmcDevice",
    "XbmcDeviceFinder",
    "VlcDevice",
    "VlcDeviceFinder"
]



import typing

import pychromecast

from . import Device, DeviceFinder
from .. import Config
from ..tools import run_method_in_executor


__all__ = [
    "ChromecastDevice",
    "ChromecastDeviceFinder"
]


class ChromecastDevice(Device):
    _device: pychromecast.Chromecast

    # noinspection PyMissingConstructor
    def __init__(self, device: typing.Any):
        self._device = device
        self._device.wait()

    def get_device_name(self) -> str:
        return self._device.device.friendly_name

    async def stop(self):
        pass

    @run_method_in_executor
    def play(self, url: str, title: str):
        self._device.media_controller.play_media(url, "video/mp4", title=title)
        self._device.media_controller.block_until_active()


class ChromecastDeviceFinder(DeviceFinder):
    @staticmethod
    @run_method_in_executor
    def find(config: Config) -> typing.List[Device]:
        return [
            ChromecastDevice(device)
            for device in pychromecast.get_chromecasts(
                timeout=config.chromecast_scan_timeout)
        ]
